import time
from .. import can

class LogPlayer:
    running = False
    def __init__(self, log_filename):
        self.logfile = open(log_filename, 'r')

    def start(self):
        assert not self.running, 'cannot start, already running'
        self.start_timestamp = time.time()
        self.running = True

    def recv(self):
        assert self.running, 'not running'
        last_timestamp = 0
        line = self.logfile.readline()
        if line == '':
            return None

        # convert line to frame
        frame = self._log_to_frame(line)

        # sleep until message occurs
        time.sleep(max((self.start_timestamp - time.time() + frame.timestamp),
                        0))

        return frame

    def _log_to_frame(self, line):
        fields = line.split(' ')

        id = int(fields[1], 0)
        frame = can.Frame(id)

        frame.timestamp = float(fields[0])

        if fields[2].upper() == 'D':
            frame.frame_type = can.FrameType.DataFrame
        elif fields[2].upper() == 'R':
            frame.frame_type = can.FrameType.RemoteFrame
        elif fields[2].upper() == 'E':
            frame.frame_type = can.FrameType.ErrorFrame
        elif fields[2].upper() == 'O':
            frame.frame_type = can.FrameType.OverloadFrame
        else:
            raise ValueError('invalid frame type')

        frame.dlc = int(fields[3])
        data = [int(fields[4], 0), int(fields[5], 0), int(fields[6], 0),
                int(fields[7], 0), int(fields[8], 0), int(fields[9], 0),
                int(fields[10], 0), int(fields[11], 0)]
        frame.data = data

        return frame



from . import can

class Bus:
    # message that belong to this bus
    _messages = []

    def add_message(self, message):
        assert isinstance(message, Message), 'invalid message'
        if message in self._messages:
            raise ValueError('Message %s already in bus' % message)
        else:
            self._messages.append(message)

    def remove_message(self, message):
        assert isinstance(Message, message), 'invalid message'
        try:
            self._messages.remove(message)
        except ValueError:
            raise ValueError('Message %s is not in bus' % message)

    def parse_frame(self, frame):
        assert isinstance(frame, can.Frame), 'invalid frame'
        for message in self._messages:
            if message.id == frame.id:
                return message.parse_frame(frame)

    def __str__(self):
        s = "Bus:\n"
        for message in self._messages:
            s = s + message.__str__()
        return s

class Message(object):
    # signals that belong to this message, indexed by start bit
    _signals = {}

    def __init__(self, name, id):
        self.name = name
        self.id = id

    def add_signal(self, signal, start_bit):
        assert isinstance(signal, Signal), 'invalid signal'
        assert isinstance(start_bit, int) and start_bit < 63, 'invalid start bit'
        self._signals[start_bit] = signal

    def remove_signal(self, signal):
        pass

    def parse_frame(self, frame):
        assert isinstance(frame, can.Frame), 'invalid frame'
        assert frame.id == self.id, 'frame id does not match message id'

        # combine 8 data bytes into single value
        frame_value = 0
        for i in range(0, frame.dlc):
            if frame.data[i] != None:
                frame_value = frame_value + (frame.data[i] << (8 * i))

        result_signals = []

        # iterate over signals
        for start_bit, signal in self._signals.items():

            # find the last bit of the singal
            end_bit = signal.bit_length + start_bit

            # compute the mask
            mask = 0
            for j in range(start_bit, end_bit):
                mask = mask + 2**j

            # apply the mask, then downshift
            value = (frame_value & mask) >> start_bit
            # pass the maksed value to the signal
            signal.parse_value(value)

            result_signals.append(signal)

        return result_signals

    def __str__(self):
        s = "Message: %s, ID: 0x%X\n" % (self.name, self.id)
        for start_bit, signal in self._signals.items():
            s = s + "\t" + signal.__str__() + "\n"
        return s

class Signal:
    def __init__(self, name, bit_length, factor=1, offset=0):
        self.name = name
        self.bit_length = bit_length
        self.factor = factor
        self.offset = offset
        self.value = 0

    def parse_value(self, value):
        self.value = value * self.factor + self.offset
        return self

    def __str__(self):
        s = "Signal: %s\tValue = %d" % (self.name, self.value)
        return s



""" can.py

Defines the low-level implementation of CAN.

"""

class FrameType:
    """ Enumerates the types of CAN frames """
    DataFrame = 1
    RemoteFrame = 2
    ErrorFrame = 3
    OverloadFrame = 4

class Frame(object):
    """ Represents a CAN Frame

    Attributes:
        id (int): CAN identifier of the Frame
        dlc (int): data length code
        data (list of int): CAN data bytes
        frame_type (int): type of CAN frame
        dlc (int): data length code of frame
        is_extended_id (bool): is this frame an extended identifier frame?
    """

    _id = None
    _data = []
    _frame_type = None
    dlc = 0
    timestamp = None

    def __init__(self, id, dlc = 0, data = [], frame_type = FrameType.DataFrame,
                 is_extended_id = False):
        """ Initializer of Frame
        Args:
            id (int): identifier of CAN frame
            dlc (int, optional): data length code of frame (defaults to 0)
            data (list, optional): data of CAN frame, defaults to empty list
            frame_type (int, optional): type of frame, defaults to 
                                        FrameType.DataFrame
            is_extended_id (bool, optional): is the frame an extended id frame?
                                             defaults to False
        """
        self.id = id
        self.data = data
        self.frame_type = frame_type
        self.is_extended_id = is_extended_id
        self.dlc = dlc

    @property
    def id(self):
        return self._id
    @id.setter
    def id(self, value):
        # ensure value is an integer
        assert isinstance(value, int), 'id must be an integer'
        # ensure standard id is in range
        if value >= 0 and value <= 0x7FF:
            self._id = value
        # otherwise, check if frame is extended
        elif value > 0x7FF and value <= 0x1FFFFFFF and self.is_extended_id:
            self._id = value
        # otherwise, id is not valid
        else:
            raise ValueError('CAN ID out of range')

    @property
    def data(self):
        result = []
        # return bytes up to dlc length, pad with zeros
        for i in range(0, min(self.dlc, len(self._data))):
            result.append(self._data[i])
        result = result + [0] * (8 - len(result))
        return result

    @data.setter
    def data(self, value):
        # data should be a list
        assert isinstance(value, list), 'CAN data must be a list'
        # data can only be 8 bytes maximum
        assert not len(value) > 8, 'CAN data cannot contain more than 8 bytes'
        # each byte must be a valid byte, int between 0x0 and 0xFF
        for byte in value:
            assert isinstance(byte, int), 'CAN data must consist of bytes'
            assert byte >= 0 and byte <= 0xFF, 'CAN data must consist of bytes'
        # data is valid
        self._data = value

    @property
    def frame_type(self):
        return self._frame_type
    @frame_type.setter
    def frame_type(self, value):
        assert value == FrameType.DataFrame or value == FrameType.RemoteFrame \
               or value == FrameType.ErrorFrame or \
               value == FrameType.OverloadFrame, 'invalid frame type'
        self._frame_type = value

    @property
    def dlc(self):
        return self._dlc
    @dlc.setter
    def dlc(self, value):
        assert isinstance(value, int)
        assert value >= 0 and value <= 8, 'dlc must be between 0 and 8'
        self._dlc = value

    def __str__(self):
        ext_str = 'standard'
        if self.is_extended_id:
            ext_str == 'extended'
        if self.frame_type == FrameType.DataFrame:
            type_str = "data"
        elif self.frame_type == FrameType.RemoteFrame:
            type_str = "remote"
        elif self.frame_type == FrameType.ErrorFrame:
            type_str = "error"
        elif self.frame_type == FrameType.OverloadFrame:
            type_str = "overload"

        result = 'Frame:'
        result = result + '\tID=0x%X, DLC=%d, ' % (self.id, self.dlc)
        result = result + '%s, %s' % (ext_str, type_str)
        result = result + '\n\tData=%s' % self.data
        return result




import os
import struct
import socket
import time

from .. import can

class SocketCanDev:
    def __init__(self, ndev):
        self.running = False
        self.socket = socket.socket(socket.PF_CAN, socket.SOCK_RAW,
                                    socket.CAN_RAW)
        self.ndev = ndev

    def start(self):
        self.socket.bind((self.ndev,))
        self.start_time = time.time()
        self.running = True

    def recv(self):
        assert self.running, 'device not running'
        frame_format = "=IB3xBBBBBBBB"
        frame_size = struct.calcsize(frame_format)

        frame_raw = self.socket.recv(frame_size)
        id, dlc, d0, d1, d2, d3, d4, d5, d6, d7 = struct.unpack(frame_format,
                                                                frame_raw)

        frame = can.Frame(id)
        frame.dlc = dlc
        frame.data = [d0,d1,d2,d3,d4,d5,d6,d7]
        frame.timestamp = time.time() - self.start_time

        return frame

    def send(self, frame):
        assert self.running, 'device not running'
        frame_format = "=IBBBBBBBBBBBB"

        data = frame.data
        packet = struct.pack(frame_format, frame.id, frame.dlc,
                    0xff, 0xff, 0xff, data[0], data[1], data[2], data[3], 
                    data[4], data[5], data[6], data[7])
        self.socket.send(packet)



from . import can

class Logger:
    def __init__(self):
        self._buffer = []

    def log(self, frame):
        assert isinstance(frame, can.Frame), 'invalid frame'

        if frame.frame_type == can.FrameType.DataFrame:
            type_char = "D"
        elif frame.frame_type == can.FrameType.RemoteFrame:
            type_char = "R"
        elif frame.frame_type == can.FrameType.ErrorFrame:
            type_char = "E"
        elif frame.frame_type == can.FrameType.OverloadFrame:
            type_char = "O"

        line = ("%f 0x%X %s %d 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X\n" %
               (frame.timestamp, frame.id, type_char, frame.dlc, frame.data[0],
                frame.data[1], frame.data[2], frame.data[3], frame.data[4],
                frame.data[5], frame.data[6], frame.data[7]))
        self._buffer.append(line)

    def __str__(self):
        s = ""
        for line in self._buffer:
            s = s + line
        return s

    def write_file(self, filename):
        with open(filename, 'w') as logfile:
            logfile.writelines(self._buffer)

    def clear(self):
        self._buffer = []



import canard.can as can
import unittest

class CanTest(unittest.TestCase):
    def setUp(self):
        self.frame = can.Frame(0)

    def test_id(self):
        # test non-integer fails
        with self.assertRaises(AssertionError):
            self.frame.id = 'boo!'

        # test negative id fails
        with self.assertRaises(ValueError):
            self.frame.id = -1

        # test standard id out of range
        self.frame.is_extended_id = False
        with self.assertRaises(ValueError):
            self.frame.id = 0x800
        # test standard id in range
        self.frame.id = 0x7FF
        self.frame.id = 0x0

        # test extended id out of range
        self.frame.is_extended_id = True
        with self.assertRaises(ValueError):
            self.frame.id = 0x2FFFFFFF
        # test extended id in range
        self.frame.id = 0x1FFFFFFF
        self.frame.id = 0x0

    def test_data(self):
        # test too many bytes causes error
        with self.assertRaises(AssertionError):
            self.frame.data = [1,2,3,4,5,6,7,8,9]
        # test wrong datatype causes error
        with self.assertRaises(AssertionError):
            self.frame.data = 4
        # test out of range byte causes error
        with self.assertRaises(AssertionError):
            self.frame.data = [1,2,3,4,5,6,7,0xFFFF]
        # check padding
        self.frame.data = [1,2,3,4,5,6,7,8]
        self.frame.dlc = 8
        self.assertEqual(self.frame.data,
                         [1, 2, 3, 4, 5, 6, 7, 8])

        self.frame.dlc = 3
        self.assertEqual(self.frame.data,
                         [1, 2, 3, None, None, None, None, None])
        self.frame.dlc = 0
        self.assertEqual(self.frame.data,
                         [None, None, None, None, None, None, None, None])

    def test_frame_type(self):
        # test invalid frame type causes error
        with self.assertRaises(AssertionError):
            self.frame.frame_type = 5
        # test valid frame types
        self.frame.frame_type = can.FrameType.DataFrame
        self.frame.frame_type = can.FrameType.RemoteFrame
        self.frame.frame_type = can.FrameType.ErrorFrame
        self.frame.frame_type = can.FrameType.OverloadFrame

    def test_dlc(self):
        with self.assertRaises(AssertionError):
            self.frame.dlc = 'test'
        with self.assertRaises(AssertionError):
            self.frame.dlc = 9
        with self.assertRaises(AssertionError):
            self.frame.dlc = -1

    def test_init(self):
        frame = can.Frame(0x55, 6, [1,2,3,4,5,6], can.FrameType.ErrorFrame, True)
        self.assertEqual(frame.id, 0x55)
        self.assertEqual(frame.data, [1,2,3,4,5,6,None,None])
        self.assertEqual(frame.frame_type, can.FrameType.ErrorFrame)
        self.assertEqual(frame.is_extended_id, True)

if __name__ == '__main__':
    unittest.main()



from distutils.core import setup

setup(
    name='CANard',
    version='0.1.0',
    author='Eric Evenchick',
    author_email='eric@evenchick.com',
    packages=['canard', 'canard.test', 'canard.hw'],
    scripts=[],
    url='http://github.com/ericevenchick/CANard',
    license='GPLv3',
    description='Library for interacting with Controller Area Network (CAN)',
    long_description=open('README.rst').read(),
    classifiers=['Development Status :: 3 - Alpha',
		 'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
		 'Topic :: Software Development :: Libraries',
		 'Topic :: Software Development :: Embedded Systems']


)



import os
import struct
import socket
import time

from .. import can
from .socketcan import SocketCanDev

class PcanDev(SocketCanDev):
    def __init__(self, minor_number=32, ndev="can0"):
        SocketCanDev.__init__(self, ndev=ndev)
        self.device_filename = "/dev/pcan%d" % minor_number

    def _write_to_chardev(self, string):
        os.system("echo '%s' > %s" % (string, self.device_filename))

    def set_btr(self, value):
        self._write_to_chardev('i 0x%X e' % value)



import serial
import io

from .. import can

class CantactDev:
    def __init__(self, port):
	self.ser = serial.Serial(port)

    def start(self):
	self.ser.write('O\r')

    def stop(self):
	self.ser.write('C\r')

    def set_bitrate(self, bitrate):
	if bitrate == 125000:
	    self.ser.write('S0\r')
	elif bitrate == 250000:
	    self.ser.write('S1\r')
	elif bitrate == 500000:
	    self.ser.write('S2\r')
	else:
	    raise ValueError("Bitrate not supported")

    def recv(self):
	# receive characters until a newline (\r) is hit
	rx_str = ""
	while rx_str == "" or rx_str[-1] != '\r':
	    rx_str = rx_str + self.ser.read()

	# parse the id, create a frame
	frame_id = int(rx_str[1:4], 16)
	frame = can.Frame(frame_id)

	# parse the DLC
	frame.dlc = int(rx_str[4])

	# parse the data bytes
	data = []
	for i in range(0, frame.dlc):
	    data.append(int(rx_str[5+i*2:7+i*2], 16))
	frame.data = data

	return frame

    def send(self, frame):
	# add type, id, and dlc to string
	tx_str = "%s%03X%d" % ('t', frame.id, frame.dlc)

	# add data bytes to string
	for i in range(0, frame.dlc):
	    tx_str = tx_str + ("%02X" % frame.data[i])

	# add newline (\r) to string
	tx_str = tx_str + '\r'

	# send it
	self.ser.write(tx_str)
